#!/usr/bin/env python
# encoding: utf-8

"""HdfsCLI Avro: an Avro extension for HdfsCLI.

Usage:
  hdfscli-avro schema [-a ALIAS] [-v...] HDFS_PATH
  hdfscli-avro read [-a ALIAS] [-v...] [-F FREQ | -n NUM] [-p PARTS] HDFS_PATH
  hdfscli-avro write [-fa ALIAS] [-v...] [-C CODEC] [-S SCHEMA] HDFS_PATH
  hdfscli-avro -L | -h

Commands:
  schema                        Pretty print schema.
  read                          Read an Avro file from HDFS and output records
                                as JSON to standard out.
  write                         Read JSON records from standard in and
                                serialize them into a single Avro file on HDFS.

Arguments:
  HDFS_PATH                     Remote path to Avro file or directory
                                containing Avro part-files.

Options:
  -C CODEC --codec=CODEC        Compression codec. Available values are among:
                                null, deflate, snappy. [default: deflate]
  -F FREQ --freq=FREQ           Probability of sampling a record.
  -L --log                      Show path to current log file and exit.
  -S SCHEMA --schema=SCHEMA     Schema for serializing records. If not passed,
                                it will be inferred from the first record.
  -a ALIAS --alias=ALIAS        Alias of namenode to connect to.
  -f --force                    Overwrite any existing file.
  -h --help                     Show this message and exit.
  -n NUM --num=NUM              Cap number of records to output.
  -p PARTS --parts=PARTS        Part-files to read. Specify a number to
                                randomly select that many, or a comma-separated
                                list of numbers to read only these. Use a
                                number followed by a comma (e.g. `1,`) to get a
                                unique part-file. The default is to read all
                                part-files.
  -v --verbose                  Enable log output. Can be specified up to three
                                times (increasing verbosity each time).

Examples:
  hdfscli-avro schema /data/impressions.avro
  hdfscli-avro read -a dev snapshot.avro >snapshot.jsonl
  hdfscli-avro read -F 0.1 -p 2,3 clicks.avro
  hdfscli-avro write -f positives.avro <positives.jsonl -S "$(cat schema.avsc)"

"""

from . import AvroReader, AvroWriter
from ...__main__ import configure_client, parse_arg
from ...config import catch
from ...util import HdfsError
from docopt import docopt
from itertools import islice
from json import JSONEncoder, dumps, loads
from random import random
import sys


class _Encoder(JSONEncoder):

  """Custom encoder to support bytes and fixed strings.

  :param \*\*kwargs: Keyword arguments forwarded to the base constructor.

  """

  encoding = 'ISO-8859-1'

  def __init__(self, **kwargs):
    kwargs.update({
      'check_circular': False,
      'separators': (',', ':'),
    })
    if sys.version_info[0] == 2: # Python 3 removed the `encoding` kwarg.
      kwargs['encoding'] = self.encoding
    super(_Encoder, self).__init__(**kwargs)

  def default(self, obj): # pylint: disable=method-hidden
    """This should only ever be run in python 3."""
    if isinstance(obj, bytes):
      return obj.decode(self.encoding)
    return super(_Encoder, self).default(self, obj)


@catch(HdfsError)
def main(argv=None, client=None, stdin=sys.stdin, stdout=sys.stdout):
  """Entry point.

  :param argv: Arguments list.
  :param client: For testing.

  """
  args = docopt(__doc__, argv=argv)
  if not client:
    client = configure_client('hdfscli-avro', args)
  elif args['--log']:
    raise HdfsError('Logging is only available when no client is specified.')
  overwrite = args['--force']
  parts = parse_arg(args, '--parts', int, ',')
  if args['write']:
    writer = AvroWriter(
      client,
      args['HDFS_PATH'],
      overwrite=overwrite,
      schema=parse_arg(args, '--schema', loads),
      codec=args['--codec'],
    )
    with writer:
      records = (loads(line) for line in stdin)
      for record in records:
        writer.write(record)
  else:
    reader = AvroReader(client, args['HDFS_PATH'], parts=parts)
    with reader:
      if args['schema']:
        stdout.write('%s\n' % (dumps(reader.schema, indent=2), ))
      elif args['read']:
        encoder = _Encoder()
        num = parse_arg(args, '--num', int)
        freq = parse_arg(args, '--freq', float)
        if freq:
          for record in reader:
            if random() <= freq:
              stdout.write(encoder.encode(record))
              stdout.write('\n')
        else:
          for record in islice(reader, num):
            stdout.write(encoder.encode(record))
            stdout.write('\n')

if __name__ == '__main__':
  main()



#!/usr/bin/env python
# encoding: utf-8

"""Test Avro extension."""

from hdfs.util import HdfsError, temppath
from json import dumps, load, loads
from nose.plugins.skip import SkipTest
from nose.tools import eq_, ok_, raises
from util import _IntegrationTest
import os
import os.path as osp
import sys

try:
  from hdfs.ext.avro import (_SeekableReader, _infer_schema, AvroReader,
    AvroWriter)
  from hdfs.ext.avro.__main__ import main
except ImportError:
  SKIP = True
else:
  SKIP = False


class TestSeekableReader(object):

  def setup(self):
    if SKIP:
      raise SkipTest

  def test_normal_read(self):
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        writer.write('abcd')
      with open(tpath) as reader:
        sreader = _SeekableReader(reader)
        eq_(sreader.read(3), 'abc')
        eq_(sreader.read(2), 'd')
        ok_(not sreader.read(1))

  def test_buffered_read(self):
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        writer.write('abcdefghi')
      with open(tpath) as reader:
        sreader = _SeekableReader(reader, 3)
        eq_(sreader.read(1), 'a')
        eq_(sreader.read(3), 'bcd')
        sreader.seek(-3, os.SEEK_CUR)
        eq_(sreader.read(2), 'bc')
        eq_(sreader.read(6), 'defghi')
        ok_(not sreader.read(1))


class TestInferSchema(object):

  def _assert_schemas_equal(self, schema1, schema2):
    # Only works for flat records.
    ok_('fields' in schema1 and 'fields' in schema2)
    fields = lambda schema: sorted(schema['fields'], key=lambda f: f['name'])
    eq_(fields(schema1), fields(schema2))

  def test_flat_record(self):
    self._assert_schemas_equal(
      _infer_schema({'foo': 1, 'bar': 'hello'}),
      {
        'type': 'record',
        'fields': [
          {'type': 'int', 'name': 'foo'},
          {'type': 'string', 'name': 'bar'},
        ]
      }
    )

  def test_array(self):
    self._assert_schemas_equal(
      _infer_schema({'foo': 1, 'bar': ['hello']}),
      {
        'type': 'record',
        'fields': [
          {'type': 'int', 'name': 'foo'},
          {'type': {'type': 'array', 'items': 'string'}, 'name': 'bar'},
        ]
      }
    )


class _AvroIntegrationTest(_IntegrationTest):

  dpath = osp.join(osp.dirname(__file__), 'dat')
  schema = None
  records = None

  @classmethod
  def setup_class(cls):
    if SKIP:
      return
    super(_AvroIntegrationTest, cls).setup_class()
    with open(osp.join(cls.dpath, 'weather.avsc')) as reader:
      cls.schema = loads(reader.read())
    with open(osp.join(cls.dpath, 'weather.jsonl')) as reader:
      cls.records = [loads(line) for line in reader]

  @classmethod
  def _get_data_bytes(cls, fpath):
    # Get Avro bytes, skipping header (order of schema fields is undefined) and
    # sync marker. This assumes that the file can be written in a single block.
    with open(fpath, 'rb') as reader:
      reader.seek(-16, os.SEEK_END) # Sync marker always last 16 bytes.
      sync_marker = reader.read()
      reader.seek(0)
      content = reader.read()
      sync_pos = content.find(sync_marker)
      return content[sync_pos + 16:-16]


class TestRead(_AvroIntegrationTest):

  def test_read(self):
    self.client.upload('weather.avro', osp.join(self.dpath, 'weather.avro'))
    with AvroReader(self.client, 'weather.avro') as reader:
      eq_(list(reader), self.records)


class TestWriter(_AvroIntegrationTest):

  def test_write(self):
    writer = AvroWriter(
      self.client,
      'weather.avro',
      schema=self.schema,
    )
    with writer:
      for record in self.records:
        writer.write(record)
    with temppath() as tpath:
      self.client.download('weather.avro', tpath)
      eq_(
        self._get_data_bytes(osp.join(self.dpath, 'weather.avro')),
        self._get_data_bytes(tpath)
      )

  def test_write_empty(self):
    with AvroWriter(self.client, 'empty.avro', schema=self.schema):
      pass
    with AvroReader(self.client, 'empty.avro') as reader:
      eq_(reader.schema, self.schema)
      eq_(list(reader), [])

  @raises(HdfsError)
  def test_write_overwrite_error(self):
    # To check that the background `AsyncWriter` thread doesn't hang.
    self.client.makedirs('weather.avro')
    with AvroWriter(self.client, 'weather.avro', schema=self.schema) as writer:
      for record in self.records:
        writer.write(record)

  def test_infer_schema(self):
    with AvroWriter(self.client, 'weather.avro') as writer:
      for record in self.records:
        writer.write(record)
    with AvroReader(self.client, 'weather.avro') as reader:
      eq_(list(reader), self.records)


class TestMain(_AvroIntegrationTest):

  def test_schema(self):
    self.client.upload('weather.avro', osp.join(self.dpath, 'weather.avro'))
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        main(['schema', 'weather.avro'], client=self.client, stdout=writer)
      with open(tpath) as reader:
        schema = load(reader)
      eq_(self.schema, schema)

  def test_read(self):
    self.client.upload('weather.avro', osp.join(self.dpath, 'weather.avro'))
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        main(
          ['read', 'weather.avro', '--num', '2'],
          client=self.client,
          stdout=writer
        )
      with open(tpath) as reader:
        records = [loads(line) for line in reader]
      eq_(records, self.records[:2])

  def test_read_part_file(self):
    data = {
      'part-m-00000.avro': [{'name': 'jane'}, {'name': 'bob'}],
      'part-m-00001.avro': [{'name': 'john'}, {'name': 'liz'}],
    }
    for fname, records in data.items():
      with AvroWriter(self.client, 'data.avro/%s' % (fname, )) as writer:
        for record in records:
          writer.write(record)
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        main(
          ['read', 'data.avro', '--parts', '1,'],
          client=self.client,
          stdout=writer
        )
      with open(tpath) as reader:
        records = [loads(line) for line in reader]
      eq_(records, data['part-m-00001.avro'])

  def test_write(self):
    with open(osp.join(self.dpath, 'weather.jsonl')) as reader:
      main(
        [
          'write', 'weather.avro',
          '--schema', dumps(self.schema),
          '--codec', 'null',
        ],
        client=self.client,
        stdin=reader
      )
    with temppath() as tpath:
      self.client.download('weather.avro', tpath)
      eq_(
        self._get_data_bytes(tpath),
        self._get_data_bytes(osp.join(self.dpath, 'weather.avro'))
      )

  def test_write_codec(self):
    with open(osp.join(self.dpath, 'weather.jsonl')) as reader:
      main(
        [
          'write', 'weather.avro',
          '--schema', dumps(self.schema),
          '--codec', 'deflate',
        ],
        client=self.client,
        stdin=reader
      )
    # Correct content.
    with AvroReader(self.client, 'weather.avro') as reader:
      records = list(reader)
    eq_(records, self.records)
    # Different size (might not be smaller, since very small file).
    compressed_size = self.client.content('weather.avro')['length']
    uncompressed_size = osp.getsize(osp.join(self.dpath, 'weather.avro'))
    ok_(compressed_size != uncompressed_size)



#!/usr/bin/env python
# encoding: utf-8

"""Support for clusters using Kerberos_ authentication.

This extension adds a new :class:`hdfs.client.Client` subclass,
:class:`KerberosClient`, which handles authentication appropriately with
Kerberized clusters:

.. code-block:: python

  from hdfs.ext.kerberos import KerberosClient
  client = KerberosClient('http://host:port')

To expose this class to the command line interface (so that it can be used by
aliases), we add the following line inside the `global` section of
`~/.hdfscli.cfg` (or wherever our configuration file is located):

.. code-block:: cfg

  autoload.modules = hdfs.ext.kerberos

Here is what our earlier configuration would look like if we updated it to
support a Kerberized production grid:

.. code-block:: cfg

  [global]
  default.alias = dev
  autoload.modules = hdfs.ext.kerberos

  [dev.alias]
  url = http://dev.namenode:port

  [prod.alias]
  url = http://prod.namenode:port
  class = KerberosClient

.. _Kerberos: http://web.mit.edu/kerberos/

"""

from ..client import Client
from ..util import HdfsError
from requests_kerberos import HTTPKerberosAuth
from six import string_types
from threading import Lock, Semaphore
from time import sleep, time
import requests as rq
import requests_kerberos # For mutual authentication globals.


class KerberosClient(Client):

  """HDFS web client using Kerberos authentication.

  :param url: Hostname or IP address of HDFS namenode, prefixed with protocol,
    followed by WebHDFS port on namenode.
  :param mutual_auth: Whether to enforce mutual authentication or not (possible
    values: `'REQUIRED'`, `'OPTIONAL'`, `'DISABLED'`).
  :param max_concurrency: Maximum number of allowed concurrent requests. This
    is required since requests exceeding the threshold allowed by the server
    will be unable to authenticate.
  :param \*\*kwargs: Keyword arguments passed to the base class' constructor.

  To avoid replay errors, a timeout of 1 ms is enforced between requests.

  """

  _delay = 0.001 # Seconds.

  def __init__(self, url, mutual_auth='OPTIONAL', max_concurrency=1, **kwargs):
    # Note the handling of options passed in as strings to support
    # instantiation via the configuration file.
    self._lock = Lock()
    self._sem = Semaphore(int(max_concurrency))
    self._timestamp = time() - self._delay
    session = kwargs.setdefault('session', rq.Session())
    if isinstance(mutual_auth, string_types):
      try:
        _mutual_auth = getattr(requests_kerberos, mutual_auth)
      except AttributeError:
        raise HdfsError('Invalid mutual authentication type: %r', mutual_auth)
    else:
      _mutual_auth = mutual_auth
    session.auth = HTTPKerberosAuth(_mutual_auth)
    super(KerberosClient, self).__init__(url, **kwargs)

  def _request(self, method, url, **kwargs):
    """Overriden method to avoid authentication errors.

    Authentication will otherwise fail if too many concurrent requests are
    being made.

    """
    if not 'auth' in kwargs:
      # Request doesn't need to be authenticated, bypass this.
      return super(KerberosClient, self)._request(method, url, **kwargs)
    with self._sem:
      with self._lock:
        delay = self._timestamp + self._delay - time()
        if delay > 0:
          sleep(delay) # Avoid replay errors.
        self._timestamp = time()
      return super(KerberosClient, self)._request(method, url, **kwargs)



#!/usr/bin/env python
# encoding: utf-8

"""Test configuration module."""

from hdfs.client import Client
from hdfs.config import Config
from hdfs.util import HdfsError, temppath
from logging.handlers import TimedRotatingFileHandler
from nose.tools import eq_, ok_, nottest, raises
from string import Template
from util import save_config
import logging as lg
import os
import os.path as osp
import sys


class TestConfig(object):

  @nottest # TODO: Find cross-platform way to reset the environment variable.
  def test_config_path(self):
    path = os.getenv('HDFSCLI_CONFIG')
    try:
      with temppath() as tpath:
        os.environ['HDFSCLI_CONFIG'] = tpath
        with open(tpath, 'w') as writer:
          writer.write('[foo]\nbar=hello')
        eq_(Config().get('foo', 'bar'), 'hello')
    finally:
      if path:
        os['HDFSCLI_CONFIG'] = path
      else:
        del os['HDFSCLI_CONFIG']

  def _write_client_module(self, path, class_name):
    template = osp.join(osp.dirname(__file__), 'dat', 'client_template.py')
    with open(template) as reader:
      contents = Template(reader.read()).substitute({
        'class_name': class_name,
      })
    with open(path, 'w') as writer:
      writer.write(contents)

  def test_autoload_client_from_path(self):
    with temppath() as module_path:
      self._write_client_module(module_path, 'PathClient')
      with temppath() as config_path:
        config = Config(config_path)
        config.add_section(config.global_section)
        config.set(config.global_section, 'autoload.paths', module_path)
        config._autoload()
        client = Client.from_options({'url': ''}, 'PathClient')
        eq_(client.one, 1)

  @raises(SystemExit)
  def test_autoload_missing_path(self):
    with temppath() as module_path:
      with temppath() as config_path:
        config = Config(config_path)
        config.add_section(config.global_section)
        config.set(config.global_section, 'autoload.paths', module_path)
        config._autoload()

  def test_autoload_client_from_module(self):
    with temppath() as module_dpath:
      os.mkdir(module_dpath)
      sys.path.append(module_dpath)
      module_fpath = osp.join(module_dpath, 'mclient.py')
      self._write_client_module(module_fpath, 'ModuleClient')
      try:
        with temppath() as config_path:
          config = Config(config_path)
          config.add_section(config.global_section)
          config.set(config.global_section, 'autoload.modules', 'mclient')
          config._autoload()
          client = Client.from_options({'url': ''}, 'ModuleClient')
          eq_(client.one, 1)
      finally:
        sys.path.remove(module_dpath)

  def test_create_client_with_alias(self):
    with temppath() as tpath:
      config = Config(path=tpath)
      section = 'dev.alias'
      config.add_section(section)
      config.set(section, 'url', 'http://host:port')
      save_config(config)
      Config(path=tpath).get_client('dev')

  def test_create_client_with_alias_and_timeout(self):
    with temppath() as tpath:
      config = Config(path=tpath)
      section = 'dev.alias'
      config.add_section(section)
      config.set(section, 'url', 'http://host:port')
      config.set(section, 'timeout', '1')
      save_config(config)
      eq_(Config(path=tpath).get_client('dev')._timeout, 1)
      config.set(section, 'timeout', '1,2')
      save_config(config)
      eq_(Config(path=tpath).get_client('dev')._timeout, (1,2))

  @raises(HdfsError)
  def test_create_client_with_missing_alias(self):
    with temppath() as tpath:
      Config(tpath).get_client('dev')

  @raises(HdfsError)
  def test_create_client_with_no_alias_without_default(self):
    with temppath() as tpath:
      Config(tpath).get_client()

  def test_create_client_with_default_alias(self):
    with temppath() as tpath:
      config = Config(tpath)
      config.add_section(config.global_section)
      config.set(config.global_section, 'default.alias', 'dev')
      section = 'dev.alias'
      config.add_section(section)
      config.set(section, 'url', 'http://host:port')
      save_config(config)
      Config(tpath).get_client()

  def test_get_file_handler(self):
    with temppath() as tpath:
      config = Config(tpath)
      handler = config.get_log_handler('cmd')
      ok_(isinstance(handler, TimedRotatingFileHandler))

  def test_disable_file_logging(self):
    with temppath() as tpath:
      config = Config(tpath)
      config.add_section('cmd.command')
      config.set('cmd.command', 'log.disable', 'true')
      save_config(config)
      config = Config(tpath)
      handler = config.get_log_handler('cmd')
      ok_(not isinstance(handler, TimedRotatingFileHandler))



#!/usr/bin/env python
# encoding: utf-8

"""Test that the examples run correctly."""

from hdfs import Config
from imp import load_source
from nose.plugins.skip import SkipTest
from six import add_metaclass
from util import _IntegrationTest
import os
import os.path as osp


class _ExamplesType(type):

  """Metaclass generating a test for each example."""

  dpath = osp.join(osp.dirname(__file__), os.pardir, 'examples')

  def __new__(mcs, cls, bases, attrs):

    def make_test(fname):
      fpath = osp.join(mcs.dpath, fname)
      module = osp.splitext(fname)[0]

      def test(self):
        try:
          load_source(module, fpath)
        except ImportError:
          # Unmet dependency.
          raise SkipTest

      test.__name__ = 'test_%s' % (module, )
      test.__doc__ = 'Test for example %s.' % (fpath, )
      return test

    for fname in os.listdir(mcs.dpath):
      if osp.splitext(fname)[1] == '.py':
        test = make_test(fname)
        attrs[test.__name__] = test
    return super(_ExamplesType, mcs).__new__(mcs, cls, bases, attrs)


@add_metaclass(_ExamplesType)
class TestExamples(_IntegrationTest):

  """Empty since tests are injected by the metaclass."""

  _get_client = None

  @classmethod
  def setup_class(cls):
    super(TestExamples, cls).setup_class()
    cls._get_client = Config.get_client
    Config.get_client = staticmethod(lambda: cls.client)

  @classmethod
  def teardown_class(cls):
    Config.get_client = cls._get_client
    super(TestExamples, cls).teardown_class()



#!/usr/bin/env python
# encoding: utf-8

"""Test Hdfs client interactions with HDFS."""

from hdfs.util import *
from nose.tools import eq_, ok_, raises


class TestAsyncWriter(object):

  def test_basic(self):
    result = []
    def consumer(gen):
      result.append(list(gen))
    with AsyncWriter(consumer) as writer:
      writer.write(1)
      writer.write(2)
    eq_(result, [[1,2]])

  def test_multiple_writer_uses(self):
    result = []
    def consumer(gen):
      result.append(list(gen))
    writer = AsyncWriter(consumer)
    with writer:
      writer.write(1)
      writer.write(2)
    with writer:
      writer.write(3)
      writer.write(4)
    eq_(result, [[1,2],[3,4]])

  def test_multiple_consumer_uses(self):
    result = []
    def consumer(gen):
      result.append(list(gen))
    with AsyncWriter(consumer) as writer:
      writer.write(1)
      writer.write(2)
    with AsyncWriter(consumer) as writer:
      writer.write(3)
      writer.write(4)
    eq_(result, [[1,2],[3,4]])

  @raises(ValueError)
  def test_nested(self):
    result = []
    def consumer(gen):
      result.append(list(gen))
    with AsyncWriter(consumer) as _writer:
      _writer.write(1)
      with _writer as writer:
        writer.write(2)

  @raises(HdfsError)
  def test_child_error(self):
    def consumer(gen):
      for value in gen:
        if value == 2:
          raise HdfsError('Yo')
    with AsyncWriter(consumer) as writer:
      writer.write(1)
      writer.write(2)

  @raises(HdfsError)
  def test_parent_error(self):
    def consumer(gen):
      for value in gen:
        pass
    def invalid(w):
      w.write(1)
      raise HdfsError('Ya')
    with AsyncWriter(consumer) as writer:
      invalid(writer)


class TestTemppath(object):

  def test_new(self):
    with temppath() as tpath:
      ok_(not osp.exists(tpath))

  def test_cleanup(self):
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        writer.write('hi')
    ok_(not osp.exists(tpath))

  def test_dpath(self):
    with temppath() as dpath:
      os.mkdir(dpath)
      with temppath(dpath) as tpath:
        eq_(osp.dirname(tpath), dpath)



#!/usr/bin/env python
# encoding: utf-8

"""Test Kerberos extension."""

from mock import Mock
from nose.tools import eq_, nottest, ok_, raises
from threading import Lock, Thread
from time import sleep, time
import sys

try:
  import requests_kerberos
except ImportError:
  sys.modules['requests_kerberos'] = Mock()

from hdfs.ext.kerberos import KerberosClient


class TestKerberosClient(object):

  def test_concurrency(self):
    lock = Lock()
    calls = set()
    items = []

    def add_item(method, url, **kwargs):
      with lock:
        ok_(not items)
        items.append(kwargs['thread'])
      sleep(0.25)
      with lock:
        thread = items.pop()
        eq_(thread, kwargs['thread'])
        calls.add(thread)
      return Mock()

    session = Mock()
    session.request = add_item
    client = KerberosClient('http://nn', max_concurrency=1, session=session)
    args = ('POST', 'http://foo')
    t1 = Thread(
      target=client._request,
      args=args,
      kwargs={'thread': 1, 'auth': True}
    )
    t1.start()
    t2 = Thread(
      target=client._request,
      args=args,
      kwargs={'thread': 2, 'auth': True}
    )
    t2.start()
    t1.join()
    t2.join()
    eq_(calls, set([1, 2]))

  def test_concurrency_no_auth(self):
    lock = Lock()
    calls = set()
    items = []

    def add_item(method, url, **kwargs):
      thread = kwargs['thread']
      with lock:
        if items:
          other = items.pop()
          ok_(other != thread)
        else:
          items.append(thread)
      sleep(0.25)
      with lock:
        ok_(not items)
        calls.add(thread)
      return Mock()

    session = Mock()
    session.request = add_item
    client = KerberosClient('http://nn', max_concurrency=1, session=session)
    args = ('POST', 'http://foo')
    t1 = Thread(
      target=client._request,
      args=args,
      kwargs={'thread': 1, 'auth': True}
    )
    t1.start()
    t2 = Thread(target=client._request, args=args, kwargs={'thread': 2})
    t2.start()
    t1.join()
    t2.join()
    eq_(calls, set([1, 2]))



#!/usr/bin/env python
# encoding: utf-8

"""Read and write Pandas_ dataframes directly from HDFS.

.. literalinclude:: /../examples/dataframe.py

This extension requires both the `avro` extension and `pandas` to be installed.
Currently only Avro serialization is supported.

.. _Pandas: http://pandas.pydata.org/

"""

from .avro import AvroReader, AvroWriter
import pandas as pd


def read_dataframe(client, hdfs_path):
  """Read dataframe from HDFS Avro file.

  :param client: :class:`hdfs.client.Client` instance.
  :param hdfs_path: Remote path to an Avro file (potentially distributed).

  """
  with AvroReader(client, hdfs_path) as reader:
    # Hack-ish, but loading all elements in memory first to get length.
    return pd.DataFrame.from_records(list(reader))


def write_dataframe(client, hdfs_path, df, **kwargs):
  """Save dataframe to HDFS as Avro.

  :param client: :class:`hdfs.client.Client` instance.
  :param hdfs_path: Remote path where the dataframe will be stored.
  :param df: Dataframe to store.
  :param \*\*kwargs: Keyword arguments passed through to
    :class:`hdfs.ext.avro.AvroWriter`.

  """
  with AvroWriter(client, hdfs_path, **kwargs) as writer:
    for _, row in df.iterrows():
      writer.write(row.to_dict())
